#
#Licensed to the Apache Software Foundation (ASF) under one
#or more contributor license agreements.  See the NOTICE file
#distributed with this work for additional information
#regarding copyright ownership.  The ASF licenses this file
#to you under the Apache License, Version 2.0 (the
#"License"); you may not use this file except in compliance
#with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing,
#software distributed under the License is distributed on an
#"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#KIND, either express or implied.  See the License for the
#specific language governing permissions and limitations
#under the License.
#
require "logstash/inputs/base"
require "logstash/namespace"
require "logstash/environment"
require "thread"
require "stud/interval"

jar_path = File.expand_path(File.join(File.dirname(__FILE__), "../../.."))
LogStash::Environment.load_runtime_jars! File.join(jar_path, "vendor")

# Datahub output plugin
class LogStash::Inputs::Datahub < LogStash::Inputs::Base
  config_name "datahub"
  default :codec, "plain"
  # datahub access id
  config :access_id, :validate => :string, :required => true

  # datahub access key
  config :access_key, :validate => :string, :required => true

  # datahub service endpoint
  config :endpoint, :validate => :string, :required => true

  # datahub project name
  config :project_name, :validate => :string, :required => true

  # datahub topic name
  config :topic_name, :validate => :string, :required => true

  # 重试次数，-1为无限重试、0为不重试、>0表示需要有限次数
  config :retry_times, :validate => :number, :required => false, :default => -1

  # 重试周期，下一次重试的间隔，单位为秒
  config :retry_interval, :validate => :number, :required => false, :default => 3

  # 指定shard
  config :shard_ids, :validate => :array, :required => false, :default => []

  # 指定start cursor
  config :cursor, :validate => :hash, :required => false, :default => {}

  # seek from cursor time
  config :cursor_time, :validate => :hash, :required => false, :default => {}

  # 消费batch
  config :batch_limit, :validate => :number, :default => 100

  # 消费间隔
  config :interval, :validate => :number, :default => 2

  # 消费间隔
  config :pos_file, :validate => :string, :required => true

  # 压缩算法
  config :compress_method, :validate => :string, :required => false, :default => ""

  # 使用pb
  config :enable_pb, :validate => :boolean, :required => false, :default => true

  # 打印sdk debug信息
  config :print_debug_info, :validate => :boolean, :required => false, :default => false

  DatahubPackage = com.aliyun.datahub.client

  public

  def register
    begin
      @stop = false
      @auto_assign = @shard_ids.empty?

      # cusor锁
      @cursor_mutex = Mutex.new

      @consumer_threads = {}

      # meta锁
      @meta_mutex = Mutex.new
      @meta_cond = ConditionVariable.new
      @refresh_time = 0
      @schema_timestamp = 0

      @account = DatahubPackage.auth.AliyunAccount::new(@access_id, @access_key)
      @datahub_conf = DatahubPackage.common.DatahubConfig::new(@endpoint, @account, @enable_pb)
      @http_conf = DatahubPackage.http.HttpConfig::new()
      @http_conf.setDebugRequest(@print_debug_info)
      if @compress_method == "deflate"
        @http_conf.setCompressType(DatahubPackage.http.HttpConfig::CompressType::DEFLATE)
      elsif @compress_method == "lz4"
        @http_conf.setCompressType(DatahubPackage.http.HttpConfig::CompressType::LZ4)
      end

      @builder = DatahubPackage.DatahubClientBuilder.newBuilder()
      @builder.setDatahubConfig(@datahub_conf)
      @builder.setHttpConfig(@http_conf)

      @client = @builder.build()

      get_topic_schema()
      @shard_map = get_shards()

      # 前置校验参数
      check_params()
      # 读取checkpoint
      read_checkpoint()

      @logger.info "Init datahub success!"
    rescue => e
      @logger.error "Init failed!" + e.message + " " + e.backtrace.inspect.to_s
      raise e
    end
  end

  # def register

  def check_params()
    if @shard_map.empty?
      @logger.error "No active shard available, please check"
      raise "No active shard available, please check"
    end

    if !@auto_assign
      @logger.info "Config shardIds:" + @shard_ids.to_s
      for shard in 0...@shard_ids.size
        @logger.info "Checking shard:" + @shard_ids[shard]
        if !@shard_map.include?(@shard_ids[shard])
          @logger.error "Config shard_id [" + @shard_ids[shard] + "] not exists, check your config"
          raise "Config shard_id not exists, check your config"
        end
      end
    end

    @logger.info "Reading from shards:" + @shard_map.keys.to_s
    begin
      checkpoint_file = File.open(@pos_file, "a+")
      checkpoint_file.close
    rescue
      @logger.error "Config pos file is invalid, pos_file must point to a file."
      raise "Config pos file is invalid, pos_file must point to a file."
    end
  end

  def update_checkpoint()
    @logger.info "flush checkpoint:" + @cursor.to_s
    @cursor_mutex.synchronize {
      checkpoint_file = File.open(@pos_file, "w")
      @cursor.each {|key, value|
        checkpoint_file.write(key + ":" + value + "\n")
      }
      checkpoint_file.close
    }
  end

  def read_checkpoint()
    begin
      @logger.info "read checkpoint:" + @pos_file
      @cursor_mutex.synchronize {
        File.foreach(@pos_file) do |line|
          checkpoint = line.chomp
          cursor_param = checkpoint.split(":")
          if cursor_param.size != 2
            raise "Invalid checkpoint:" + checkpoint
          end
          @cursor[cursor_param[0]] = cursor_param[1]
        end
        @logger.info "recover checkpoint:" + @cursor.to_s
      }
    rescue => e
      @logger.error e.backtrace.inspect.to_s
      @logger.error "No checkpoint found or invalid, will start normally."
      raise e
    end
  end

  def update_cursor(shard_id, cursor)
    @cursor_mutex.synchronize {
      @cursor[shard_id] = cursor
    }
  end

  def get_cursor(shard_id, force = false)
    @cursor_mutex.synchronize {
      if (force || !@cursor.has_key?(shard_id) || @cursor[shard_id] == nil || @cursor[shard_id] == "")
        @logger.info thread_id + "Shard:" + shard_id + " has no checkpoint, will seek to begin."
        if (@cursor_time.has_key?(shard_id))
          if (@cursor_time[shard_id] == '-1')
            cursorRs = @client.getCursor(@project_name, @topic_name, shard_id, DatahubPackage.model.CursorType::LATEST)
          else
            cursorRs = @client.getCursor(@project_name, @topic_name, shard_id, DatahubPackage.model.CursorType::SYSTEM_TIME, @cursor_time[shard_id].to_i * 1000)
          end
        else
          cursorRs = @client.getCursor(@project_name, @topic_name, shard_id, DatahubPackage.model.CursorType::OLDEST)
        end
        @cursor[shard_id] = cursorRs.getCursor()
      end
      return @cursor[shard_id]
    }
  end

  def read_record(shard_id, queue)
    cursor = ""
    begin
      cursor = get_cursor(shard_id)
      @logger.info thread_id + "Shard: " + shard_id + " Get record at cursor:" + cursor

      fields = @schema.getFields()
      recordRs = @client.getRecords(@project_name, @topic_name, shard_id, @schema, cursor, @batch_limit)
      recordEntries = recordRs.getRecords()
      recordEntries.each do |record|
        data = Hash.new
        recordData = record.getRecordData()
        fields.each do |field|
          data[field.getName()] = recordData.getField(field.getName())
        end
        data["timestamp"] = record.getSystemTime()
        event = LogStash::Event.new("message" => data)
        decorate(event)
        queue << event
      end
      cursor = recordRs.getNextCursor()
      update_cursor(shard_id, cursor)
      @logger.info thread_id + "Shard:" + shard_id + " Next cursor:" + cursor + " GetRecords:" + (recordRs.getRecordCount()).to_s
      if (recordRs.getRecordCount() == 0)
        @logger.info thread_id + "Shard: " + shard_id + " Read to latest, waiting for data:" + cursor
        Stud.stoppable_sleep([@retry_interval, 0.5].max) {stop?}
      else
        update_checkpoint
      end
    rescue DatahubPackage.exception.DatahubClientException => e
      if e.getErrorCode().to_s == "InvalidShardOperation"
        @logger.error thread_id + "Shard:" + shard_id + " is sealed, cursor:" + cursor
        Thread.current["sealed"] = true
        @meta_mutex.synchronize {
          @meta_cond.signal
        }
        Stud.stoppable_sleep(@retry_interval) {stop?}
      elsif e.getErrorCode().to_s == "NoSuchShard"
        @logger.error thread_id + "Shard:" + shard_id + " is not found, cursor:" + cursor
        @meta_mutex.synchronize {
          @meta_cond.signal
        }
        Stud.stoppable_sleep(@retry_interval) {stop?}
      elsif e.getErrorCode().to_s == "InvalidCursor"
        @logger.error thread_id + "Shard:" + shard_id + " Invalid cursor:" + cursor + ", seek to begin."
        get_cursor(shard_id, true)
      else
        @logger.error thread_id + "Read failed:" + e.getMessage() + ", will retry later."
        Stud.stoppable_sleep(@retry_interval) {stop?}
      end
    end
  end

  def shard_consumer(shard_id, queue)
    @logger.info thread_id + "Consumer thread started for shard:" + shard_id
    Thread.current["sealed"] = false
    Thread.current["closed"] = false
    while !check_stop && !Thread.current["closed"] do
      begin
        if !@shard_map.include?(shard_id) ||
            (@shard_map[shard_id].getState() == DatahubPackage.model.ShardState::CLOSED && Thread.current["sealed"])
          @logger.error thread_id + "Shard:" + shard_id + " is sealed and read to end, sleep and retry"
          Stud.stoppable_sleep([@retry_interval, 5].max) {stop?}
        else
          read_record(shard_id, queue)
        end
      rescue => e
        @logger.error thread_id + "Read records failed" + e.backtrace.inspect.to_s
        @logger.error thread_id + "Will retry after " + @retry_interval.to_s + " seconds."
        Stud.stoppable_sleep(@retry_interval) {stop?}
      end
    end # loop
    @logger.info thread_id + "Consumer thread exit, shard:" + shard_id
  end

  def check_stop()
    @cursor_mutex.synchronize {
      return @stop
    }
  end

  def run(queue)
    if @auto_assign
      @shard_map.each_key do |shard_id|
        @consumer_threads[shard_id] = Thread.new(shard_id, queue) {|shard_id, queue|
          shard_consumer(shard_id, queue)
        }
      end
    else
      @shard_ids.each do |shard_id|
        @consumer_threads[shard_id] = Thread.new(shard_id, queue) {|shard_id, queue|
          shard_consumer(shard_id, queue)
        }
      end
    end

    while !check_stop do
      @meta_mutex.synchronize {
        @meta_cond.wait(@meta_mutex, 300)
      }

      if check_stop
        break
      end
      @logger.debug thread_id + "Main thread heartbeat."

      now = Time.now.getutc.to_i

      if @schema_timestamp + 300 < now
        @logger.info thread_id + "refresh topic schema " + now.to_s
        get_topic_schema
        @schema_timestamp = now
      end

      if @refresh_time + 5 < now
        @logger.info thread_id + "update shard list " + now.to_s
        @shard_map = get_shards()
      end

      @shard_map.each_key do |shard_id|
        if !@consumer_threads.include?(shard_id)
          @consumer_threads[shard_id] = Thread.new(shard_id, queue) {|shard_id, queue|
            shard_consumer(shard_id, queue)
          }
        end
      end

      @consumer_threads.each do |shard_id, consumer|
        if !@shard_map.include?(shard_id)
          consumer["closed"] = true
          @logger.info thread_id + "Shard:" + shard_id + " is recycled, remove consumer"
        end
      end
      @consumer_threads.delete_if do |shard_id, _|
        !@shard_map.include?(shard_id)
      end
    end # loop

    @logger.info thread_id + "Plugin stopping, waiting for consumer thread stop."
    @consumer_threads.each_value {|thread|
      thread.join
    }
    @logger.info thread_id + "Plugin stopped."
  end

  # def run

  def get_shards()
    shards = @client.listShard(@project_name, @topic_name).getShards()
    shardMap = {}
    shards.each do |shard|
      shardMap[shard.getShardId()] = shard
    end
    return shardMap
  end

  def get_topic_schema()
    topic_meta = @client.getTopic(@project_name, @topic_name)
    @schema = topic_meta.getRecordSchema()
    fields = @schema.getFields()
    columns_size = fields.size
    column_names = []
    for i in 0...columns_size
      column_names.push(fields[i].getName())
    end
    @logger.info thread_id + "schema: " + column_names.to_s
  end

  def stop
    @cursor_mutex.synchronize {
      @stop = true
    }
    @meta_mutex.synchronize {
      @meta_cond.signal
    }
  end

  def thread_id
    "[" + Thread.current.object_id.to_s + "]"
  end

end # class LogStash::Inputs::Datahub
